#!/usr/bin/env python

# http://iki.fi/teknohog/hacks/idler/

# idler.py: spindown daemon for external drives, as they cannot be
# controlled via hdparm -S. Idea adapted from rusi_sg_down, added
# threading for multiple devices. Further ideas adapted from Spindown.

# release 20090315

# requires sys-apps/sg3_utils

# Copyright (C) 2009 Risto A. Paju
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from os.path import join, exists, realpath, basename
from os import spawnv, P_WAIT, environ
from ConfigParser import SafeConfigParser
from time import time, sleep
from threading import Thread

config = SafeConfigParser()
config.read(environ["IDLER_CFG"])

sg_path = config.get("general", "sg_path")

class Idlette(Thread):
    def __init__(self, drive):
        Thread.__init__(self)

        # individual timeouts are preferred to general
        if "timeout" in config.options(drive):
            self.timeout = int(config.get(drive, "timeout"))
        elif "timeout" in config.options("general"):
            self.timeout = int(config.get("general", "timeout"))

        # id-based definition is preferred to device-based

        # use basename for extra safety, so that we accept
        # either sda or /dev/sda
        if "id" in config.options(drive):
            self.id = basename(config.get(drive, "id"))
            self.devpath = join("/dev/disk/by-id/", self.id)
        elif "device" in config.options(drive):
            self.device = basename(config.get(drive, "device"))
            self.devpath = join("/dev/", self.device)

        self.request_count_old = 0
        self.bedtime = 0
        self.last_access_time = 0
        self.sg_args = [sg_path, "--stop", self.devpath]

    def run(self):
        while True:
            # the device may not exist yet, but start the loop anyway,
            # in case the device comes up

            # note: os.path.isfile does not work with device files, or
            # with symlinks to them. use os.path.exists instead.
            if exists(self.devpath):
                # For id-based drives, we can only resolve the real
                # device file path at this stage. This is somewhat
                # roundabout for device-defined drives. Of course, we
                # could write somewhat different loops for each case,
                # but that would require some ifs, extra function
                # calls, and/or repeated code, so it would not
                # necessarily be any better.

                self.device = basename(realpath(self.devpath))
                self.iorq_path = join('/sys/block', self.device, 'device/iorequest_cnt')
                
                f = open(self.iorq_path, 'r')
                self.request_count = int(f.read(), 16)
                f.close()

                self.now = time()
                
                if self.request_count != self.request_count_old:
                    self.last_access_time = self.now
                # after one spindown command, the drive stays idle,
                # and we should not issue further commands. thus no
                # commands between bedtime and waking up :)
                elif self.now - self.last_access_time > self.timeout \
                         and self.last_access_time > self.bedtime:
                    spawnv(P_WAIT, self.sg_args[0], self.sg_args)
                    self.bedtime = self.now
                    # increase the request count by one for the command we just sent
                    self.request_count += 1
                
                self.request_count_old = self.request_count

            sleep(self.timeout / 10.)

id_threads = []
for drive in config.sections():
    if drive != "general":
        id_threads.append(Idlette(drive))
        id_threads[-1].start()
